#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
import argparse
import asyncio
import logging
import sys
import typing
import nanaimo


class _ArgparseSubparserArguments(nanaimo.Arguments):
    """
    Nanaimo :class:`Arguments` that delegates to a wrapped
    :class:`argparse.ArgumentParser` instance.
    """

    @classmethod
    def visit_argparse(cls,
                       manager: nanaimo.FixtureManager,
                       subparsers: argparse._SubParsersAction,
                       loop: asyncio.AbstractEventLoop) -> None:
        for fixture_type in manager.fixture_types():
            subparser = subparsers.add_parser(fixture_type.get_canonical_name())  # type: 'argparse.ArgumentParser'
            subparser.add_argument('--test-timeout-seconds',
                                   default='30',
                                   type=float,
                                   help='''Test will be killed and marked as a failure after
waiting for a result for this amount of time.''')
            fixture_type.on_visit_test_arguments(cls(subparser))

            async def create_and_gather_on_default(args: nanaimo.Namespace) -> int:
                """
                Stores the type, manager, and loop then uses these to instantiate
                and invoke a Fixture if the given default is selected.
                Returns the result-code of the artifacts.
                """
                nonlocal manager, loop
                fixture = fixture_type(manager, loop)
                return int(await fixture.gather(args))

            subparser.set_defaults(func=create_and_gather_on_default)

    def __init__(self, subparser: argparse.ArgumentParser):
        self._subparser = subparser

    def add_argument(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        self._subparser.add_argument(*args, **kwargs)

    def set_defaults(self, **kwargs: typing.Any) -> None:
        self._subparser.set_defaults(**kwargs)


def _make_parser(loop: typing.Optional[asyncio.AbstractEventLoop] = None) -> argparse.ArgumentParser:
    """
    Defines the command-line interface. Provided as a separate factory method to
    support sphinx-argparse documentation.
    """

    epilog = '''**Example Usage**::

    python -m nanaimo -vv nanaimo_bar

----
'''

    parser = argparse.ArgumentParser(
        description='Run tests against hardware.',
        epilog=epilog,
        formatter_class=argparse.RawTextHelpFormatter)

    from nanaimo.version import __version__

    parser.add_argument('--version', action='version', version=__version__)

    parser.add_argument('--verbose', '-v', action='count',
                        help='verbosity level (-v, -vv)')

    subparsers = parser.add_subparsers(help='sub-command help')

    pm = nanaimo.FixtureManager()

    _ArgparseSubparserArguments.visit_argparse(pm, subparsers, loop)  # type: ignore

    return parser


def _setup_logging(args: argparse.Namespace) -> None:
    fmt = '%(name)s : %(message)s'
    level = {0: logging.WARNING, 1: logging.INFO,
             2: logging.DEBUG}.get(args.verbose or 0, logging.DEBUG)
    logging.basicConfig(stream=sys.stderr, level=level, format=fmt)
    if args.verbose is not None and args.verbose >= 3:
        logging.getLogger('asyncio').setLevel(logging.DEBUG)


def main() -> int:
    """
    CLI entrypoint for running Nanaimo tests.
    """

    loop = asyncio.get_event_loop()

    parser = _make_parser(loop)
    args = parser.parse_args()

    _setup_logging(args)

    if hasattr(args, 'func'):
        result = loop.run_until_complete(args.func(nanaimo.Namespace(args)))
        try:
            return int(result)
        except ValueError:
            print('Nanaimo tests must return an int result!')
            raise
    else:
        parser.print_usage()
        return -1



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
"""
Nanaimo presents itself as a single pytest fixture called `nanaimo_fixture_manager`
which allows tests to access or control test hardware fixtures. To register your
test fixture with the Nanaimo fixture manager use the `nanaimo.FixtureManager.type_factory`
:class:`pluggy.HookimplMarker` to register your :class:`nanaimo.Fixture`.

.. invisible-code-block: python
    import nanaimo
    import pytest
    import typing

.. code-block:: python

    # In my_namespace/__init__.py

    class MyTestFixture(nanaimo.Fixture):

        @classmethod
        def on_visit_test_arguments(cls, arguments: nanaimo.Arguments) -> None:
            pass

        async def gather(self, args: nanaimo.Namespace) -> nanaimo.Artifacts:
            artifacts = nanaimo.Artifacts()
            # Do your on-target testing here and store results in nanaimo.Artifacts.
            return artifacts


    @nanaimo.FixtureManager.type_factory
    def get_fixture_type() -> typing.Type['nanaimo.Fixture']:
        return MyTestFixture

Individual fixtures can choose to present pytest fixtures directly using `pytest.fixture`
and :func:`create_pytest_fixture`

.. code-block:: python

    @pytest.fixture
    def my_test_fixture(request: typing.Any) -> nanaimo.Fixture:
        return nanaimo.pytest_plugin.create_pytest_fixture(request, MyTestFixture)


In your setup.cfg you'll first need to register nanaimo with pytest ::

    [options]
        pytest11 =
            pytest_nanaimo = nanaimo.pytest_plugin

then, if you do want to expose your fixture directly, you'll need to add your fixture's namespace::

    [options]
        pytest11 =
            pytest_nanaimo = nanaimo.pytest_plugin
            pytest_my_plugin = my_namspace

"""
import typing

import pytest

import nanaimo

_fixture_manager = None  # type: typing.Optional[nanaimo.FixtureManager]
"""
Pytest plugin singleton. The first time our pytest plugin is invoked
we populate to provide a peristent store of Nanaimo fixtures and artifacts.
"""


def _get_default_fixture_manager() -> nanaimo.FixtureManager:
    global _fixture_manager
    if _fixture_manager is None:
        _fixture_manager = nanaimo.FixtureManager()
    return _fixture_manager


def create_pytest_fixture(pytest_request: typing.Any, fixture_type: typing.Type[nanaimo.Fixture]) -> nanaimo.Fixture:
    """
    Create a fixture for a `pytest.fixture` request. This method ensures the fixture is created through
    the default :class:`FixtureManager`.

    :param pytest_request: The request object passed into the pytest fixture factory.
    :type pytest_request: _pytest.fixtures.FixtureRequest
    :return: Either a new fixture or a fixture of the same name that was already created for the default
        :class:`FixtureManager`.
    """
    return _get_default_fixture_manager().get_fixture(fixture_type.get_canonical_name())


class _PyTestArguments(nanaimo.Arguments):

    def __init__(self, testparser):  # type: ignore
        self._testparser = testparser

    def add_argument(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        self._testparser.addoption(*args, **kwargs)

    def set_defaults(self, **kwargs: typing.Any) -> None:
        raise NotImplementedError('pytest plugin does not support setting defaults.')


def pytest_addoption(parser) -> None:  # type: ignore
    manager = _get_default_fixture_manager()
    for fixture_type in manager.fixture_types():
        group = parser.getgroup(fixture_type.get_canonical_name())
        args = _PyTestArguments(group)
        fixture_type.on_visit_test_arguments(args)


@pytest.fixture
def nanaimo_fixture_manager(request: typing.Any) -> nanaimo.FixtureManager:
    """
    The default fixture for Nanaimo. Available as `nanaimo_fixture_manager`

    .. code-block:: python

        def test_example(nanaimo_fixture_manager: nanaimo.FixtureManager) -> None:
            my_fixture = nanaimo_fixture_manager.get_fixture('my_fixture')

    :param pytest_request: The request object passed into the pytest fixture factory.
    :type pytest_request: _pytest.fixtures.FixtureRequest
    :return: A reference to the default fixture manager.
    """
    return _get_default_fixture_manager()



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
import nanaimo.version
import nanaimo.connections
import nanaimo.instruments.bkprecision
import asyncio
import argparse
import sys
import logging


def _make_parser() -> argparse.ArgumentParser:
    epilog = '''**Example Usage**::

    python -m nanaimo.instruments.bkprecision \\
        --port /dev/serial/by-id/usb_TTL232R_BLAH_port0 \\
        on

----
'''

    parser = argparse.ArgumentParser(
        description='Send one-shot commands to the bk-precision',
        epilog=epilog,
        formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('--version', action='version', version=nanaimo.version.__version__)

    parser.add_argument('--verbose', '-v', action='count',
                        help='verbosity level (-v, -vv)')

    parser.add_argument('--local-echo', '-e', action='store_true',
                        help='Enable debug logging of sent commands. Use with -vv.')

    parser.add_argument('--port',
                        help='The port the BK Precision power supply is connected to.')

    parser.add_argument('--timeout-seconds', default=2.5,
                        help='fractional seconds to wait for an OK before marking the command as failed.')

    parser.add_argument('command', help='Command to send to the power supply.')
    return parser


args = _make_parser().parse_args()

#
# Setup Python logging.
#
fmt = '%(name)20s %(levelname)7s: %(message)s'
level = {0: logging.WARNING, 1: logging.INFO,
         2: logging.DEBUG}.get(args.verbose or 0, logging.DEBUG)
logging.basicConfig(stream=sys.stderr, level=level, format=fmt)

#
# Start asyncio loop (3.5 compatible version)
#
loop = asyncio.get_event_loop()

with nanaimo.connections.uart.ConcurrentUart.new_default(args.port, 9600, loop) as serial:
    serial.echo = args.local_echo
    bk = nanaimo.instruments.bkprecision.Series1900BUart(serial, args.timeout_seconds)
    result = loop.run_until_complete(bk.send_command(args.command))

sys.exit(result)



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
import asyncio
import logging
import re
import typing

import nanaimo.connections


class Parser:
    """
    Uses a given monitor to watch for google test results.
    """

    def __init__(self, timeout_seconds: float, loop: typing.Optional[asyncio.AbstractEventLoop] = None):
        self._loop = (loop if loop is not None else asyncio.get_event_loop())
        self._logger = logging.getLogger(__name__)
        self._timeout_seconds = timeout_seconds
        self._completion_pattern = re.compile(r'\[\s*(PASSED|FAILED)\s*\]\s*(\d+)\s+tests?\.')

    async def read_test(self, uart: nanaimo.connections.AbstractAsyncSerial) -> int:
        start_time = self._loop.time()
        result = 1
        line_count = 0
        while True:
            now = self._loop.time()
            if now - start_time > self._timeout_seconds:
                result = 2
                break
            try:
                line = await uart.get_line(timeout_seconds=self._timeout_seconds - (now - start_time))
            except asyncio.TimeoutError:
                result = 2
                break
            self._logger.debug(line)
            line_count += 1
            line_match = self._completion_pattern.match(line)
            if line_match is not None:
                result = (0 if line_match.group(1) == 'PASSED' else 1)
                break
        if 0 == result:
            self._logger.info('Detected successful test after %f seconds.', self._loop.time() - start_time)
        elif 2 == result:
            self._logger.warning('gtest.Parser timeout after %f seconds', self._loop.time() - start_time)
        self._logger.debug('Processed %d lines. There were %d buffer full events reported.', line_count, uart.rx_buffer_overflows)
        return result



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
"""
TODO: Use https://python-saleae.readthedocs.io/en/latest/ to create captures which
can be presented to pytest as a fixture.
"""



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
import nanaimo
import asyncio
import logging
import pathlib
import typing


class ProgramUploaderJLink:
    """
    Async manager of a JLinkExe subprocess.
    """

    @classmethod
    def on_visit_test_arguments(cls, arguments: nanaimo.Arguments) -> None:
        arguments.add_argument('--base_path',
                               default=str(pathlib.Path().cwd()),
                               help='The folder under which to search for jlink scripts.')
        arguments.add_argument('--jlink-scripts',
                               help='A globbing pattern to collect jlink scripts for flashing tests.')
        arguments.add_argument('--upload-timeout-seconds',
                               default='20',
                               type=float,
                               help='''The upload will be killed and an error returned
after waiting for the upload to complete for this
amount of time.''')

    def __init__(self,
                 jlink_executable: pathlib.Path = pathlib.Path('JLinkExe'),
                 extra_arguments: typing.Optional[typing.List[str]] = None):
        self._logger = logging.getLogger(__name__)
        self._jlink_exe = jlink_executable
        self._extra_arguments = extra_arguments

    async def upload(self, jlink_script: pathlib.Path) -> int:
        cmd = '{} -CommanderScript {}'.format(self._jlink_exe, jlink_script)
        if self._extra_arguments is not None:
            cmd += ' ' + ' '.join(self._extra_arguments)

        self._logger.info('starting upload: %s', cmd)
        proc = await asyncio.create_subprocess_shell(
            cmd,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE
        )  # type: asyncio.subprocess.Process

        stdout, stderr = await proc.communicate()

        self._logger.info('%s exited with %i', cmd, proc.returncode)

        if stdout:
            self._logger.debug(stdout.decode())
        if stderr:
            self._logger.error(stderr.decode())

        return proc.returncode



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#

import sys

import nanaimo.cli

sys.exit(nanaimo.cli.main())



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#

import nanaimo
import pytest


def test_create_orphan() -> None:
    """
    Create an orphan Namespace
    """
    subject = nanaimo.Namespace()
    assert 'foo' not in subject


def test_create_single_parent() -> None:
    """
    Create a namespace with a single parent.
    """
    parent = nanaimo.Namespace()
    setattr(parent, 'foo', 1)
    child = nanaimo.Namespace(parent)
    assert 1 == child.foo
    with pytest.raises(KeyError):
        child.bar


def test_create_grandparent() -> None:
    """
    Create a namespace with a grand parent.
    """
    grandparent = nanaimo.Namespace()
    setattr(grandparent, 'foo', 1)
    setattr(grandparent, 'bar', 8)
    parent = nanaimo.Namespace(grandparent)
    setattr(parent, 'foo', 2)
    child = nanaimo.Namespace(parent)
    assert 2 == child.foo
    assert 8 == child.bar
    assert 1 == grandparent.foo
    assert 8 == child.bar



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
"""
The parser module contains parsers for common reporting formats used by
instruments or test frameworks.  These should be broken out from instruments
where appropriate to ensure reuse accross instrument families and communication
busses.
"""



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
"""
Nanaimo is composed of three primary object types:

    1. :class:`Fixture` – A fixture
    2. :class:`Instrument` – An Instrument
    3. :class:`Activity` – An activity

.. figure:: https://thirtytwobits.github.io/nanaimo/images/static_state_1.png
   :alt: Static-state diagram of Nanaimo's primary types.

   Static-state diagram of Nanaimo's primary types.

.. figure:: https://thirtytwobits.github.io/nanaimo/images/object_1.png
   :alt: Object example of Nanaimo's primary types in use.

   Object example of Nanaimo's primary types in use.

"""
import abc
import asyncio
import logging
import math
import typing

import pluggy


class AssertionError(RuntimeError):
    """
    Thrown by Nanaimo tests when an assertion has failed.

    .. Note::
        This exception should be used only when the state of a :class:`Fixture`
        was invalid. You should use pytest tests and assertions when writing validation
        cases for fixture output like log files or sensor data.
    """
    pass


class Arguments(metaclass=abc.ABCMeta):
    """
    Protocol for argument type that supports both argparse and pytest arguments concepts.

    .. Note::
        This will go away at some disant point in the future where Nanaimo supports only
        Python 3.8 and newer since `PEP-544 protocols <https://www.python.org/dev/peps/pep-0544/>`_
        can be used.
    """
    @abc.abstractmethod
    def add_argument(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        ...

    @abc.abstractmethod
    def set_defaults(self, **kwargs: typing.Any) -> None:
        ...


class Namespace:
    """
    Generic object that acts like :class:`argparse.Namespace` but can be created using pytest
    plugin arguments as well.
    """

    def __init__(self, parent: typing.Optional[typing.Any] = None):
        self._parent = parent

    def __getattr__(self, key: str) -> typing.Any:
        try:
            return self.__dict__[key]
        except KeyError:
            if self._parent is None:
                raise
        return getattr(self._parent, key)

    def __contains__(self, key: str) -> typing.Any:
        if key in self.__dict__:
            return True
        if self._parent is None:
            return False
        else:
            return key in self._parent


class Artifacts(Namespace):
    """
    Namespace returned by :class:`Fixture` objects when invoked that contains the artifacts collected
    from the fixture's activities.

    :param result_code: The value to report as the status of the activity that gathered the artifacts.
    :param parent: A parent namespace to lookup in if the current namespace doesn't have a value.
    """
    def __init__(self, result_code: int = 0, parent: typing.Optional[typing.Any] = None):
        super().__init__(parent)
        self._result_code = result_code

    @property
    def result_code(self) -> int:
        """
        The overall status of the fixture activities. 0 is successful all other values are
        errors.
        """
        return self._result_code

    @result_code.setter
    def result_code(self, new_result: int) -> None:
        self._result_code = new_result

    def __int__(self) -> int:
        return self._result_code


class Fixture(metaclass=abc.ABCMeta):
    """
    Common, abstract class for pytest fixtures based on Nanaimo. Nanaimo fixtures provide a visitor pattern for arguments that
    are common for both pytest extra arguments and for argparse commandline arguments. This allows a Nanaimo fixture to expose
    a direct invocation mode for debuging with the same arguments used by the fixture as a pytest plugin. Additionally all
    Nanaimo fixtures provide a :func:`gather` function that takes a :class:`Namespace` containing the provided arguments and
    returns a set of :class:`Artifacts` gathered by the fixture. The contents of these artifacts are documented by each
    concrete fixture.

    .. invisible-code-block: python
        import nanaimo
        import asyncio

        _doc_loop = asyncio.new_event_loop()

    .. code-block:: python

        class MyFixture(nanaimo.Fixture):
            @classmethod
            def on_visit_test_arguments(cls, arguments: nanaimo.Arguments) -> None:
                arguments.add_argument('--foo', default='bar')

            async def gather(self, args: nanaimo.Namespace) -> nanaimo.Artifacts:
                artifacts = nanaimo.Artifacts(-1)
                # do something and then return
                artifacts.result_code = 0
                return artifacts

    .. invisible-code-block: python
        foo = MyFixture(None)
        ns = nanaimo.Namespace()

        _doc_loop.run_until_complete(foo.gather(ns))

    `MyFixture` can now be used from a commandline like::

        python -m nanaimo MyFixture --foo baz

    or as part of a pytest::

        pytest --foo=baz

    """

    @classmethod
    def get_canonical_name(cls) -> str:
        """
        The name to use as a key for this :class:`Fixture` type.
        """
        return str(getattr(cls, 'fixture_name', cls.__name__))

    @classmethod
    @abc.abstractmethod
    def on_visit_test_arguments(cls, arguments: Arguments) -> None:
        """
        Called by the environment before instantiating any :class:`Fixture` instances to register
        arguments supported by each type. These arguments should be portable between both :mod:`argparse`
        and :mod:`pytest`. The fixture is registered for this callback by returning a reference to its
        type from a :attr:`Fixture.Manager.type_factory` annotated function registered as an entrypoint in the Python application.
        """
        ...

    def __init__(self, manager: 'FixtureManager', loop: typing.Optional[asyncio.AbstractEventLoop] = None):
        self._manager = manager
        self._name = self.get_canonical_name()
        self._logger = logging.getLogger(self._name)
        self._loop = loop

    @property
    def name(self) -> str:
        """
        The canonical name for the Fixture.
        """
        return self._name

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        """
        The running asyncio EventLoop in use by this Fixture.
        This will be the loop provided to the fixture in the constructor if that loop is still
        running otherwise the loop will be a running loop retrieved by :func:`asyncio.get_event_loop`.
        :raises RuntimeError: if no running event loop could be found.
        """
        if self._loop is None or not self._loop.is_running():
            self._loop = asyncio.get_event_loop()
        if not self._loop.is_running():
            raise RuntimeError('No running event loop was found!')
        return self._loop

    @property
    def manager(self) -> 'FixtureManager':
        """
        The :class:`FixtureManager` that owns this :class:`Fixture`.
        """
        return self._manager

    @property
    def logger(self) -> logging.Logger:
        """
        A logger for this :class:`Fixture` instance.
        """
        return self._logger

    @abc.abstractmethod
    async def gather(self, args: Namespace) -> Artifacts:
        """
        Coroutine awaited to gather fixture artifacts. The fixture should always retrieve new artifacts when invoked
        leaving caching to the caller.
        :param args:    The arguments provided for the fixture instance.
        :type args: Namespace
        :return: A set of artifacts with the :attr:`Artifacts.result_code` set to indicate the success or failure
            of the fixture's artifact gathering activies.
        """
        ...

    async def countdown_sleep(self, sleep_time_seconds: float) -> None:
        """
        Calls :func:`asyncio.sleep` for 1 second then emits an :meth:`logging.Logger.info`
        of the time remaining until `sleep_time_seconds`.
        This is useful for long waits as an indication that the process is not deadlocked.

        :param sleep_time_seconds:  The amount of time in seconds for this coroutine to wait
            before exiting. For each second that passes while waiting for this amount of time
            the coroutine will :func:`asyncio.sleep` for 1 second.
        :type sleep_time_seconds: float
        """
        count_down = sleep_time_seconds
        while count_down >= 0:
            self._logger.info('%d', math.ceil(count_down))
            await asyncio.sleep(1)
            count_down -= 1

    async def observe_tasks_assert_not_done(self,
                                            observer_co_or_f: typing.Union[typing.Coroutine, asyncio.Future],
                                            timeout_seconds: float,
                                            *args: typing.Union[typing.Coroutine, asyncio.Future]) -> typing.Set[asyncio.Future]:
        """
        Allows running a set of tasks but returning when an observer task completes. This allows a pattern where
        a single task is evaluating the side-effects of other tasks as a gate to continuing the test.
        """
        done, pending = await self._observe_tasks(observer_co_or_f, timeout_seconds, *args)
        if len(done) > 1:
            raise AssertionError('Tasks under observation completed before the observation was complete.')
        return pending

    async def observe_tasks(self,
                            observer_co_or_f: typing.Union[typing.Coroutine, asyncio.Future],
                            timeout_seconds: float,
                            *args: typing.Union[typing.Coroutine, asyncio.Future]) -> typing.Set[asyncio.Future]:
        """
        Allows running a set of tasks but returning when an observer task completes. This allows a pattern where
        a single task is evaluating the side-effects of other tasks as a gate to continuing the test.
        """

        done, pending = await self._observe_tasks(observer_co_or_f, timeout_seconds, *args)
        return pending

    # +-----------------------------------------------------------------------+
    # | PRIVATE
    # +-----------------------------------------------------------------------+

    async def _observe_tasks(self,
                             observer_co_or_f: typing.Union[typing.Coroutine, asyncio.Future],
                             timeout_seconds: float,
                             *args: typing.Union[typing.Coroutine, asyncio.Future]) -> \
            typing.Tuple[typing.Set[asyncio.Future], typing.Set[asyncio.Future]]:

        observing_my_future = asyncio.ensure_future(observer_co_or_f)
        the_children_are_our_futures = [observing_my_future]
        for co_or_f in args:
            the_children_are_our_futures.append(asyncio.ensure_future(co_or_f))

        start_time = self.loop.time()
        wait_timeout = (timeout_seconds if timeout_seconds > 0 else None)

        while True:
            done, pending = await asyncio.wait(
                the_children_are_our_futures,
                timeout=wait_timeout,
                return_when=asyncio.FIRST_COMPLETED)

            if observing_my_future.done():
                return done, pending

            if wait_timeout is not None and self.loop.time() - start_time > wait_timeout:
                break

        raise asyncio.TimeoutError()


class FixtureManager:
    """
    Object that scopes a set of :class:`Fixture`. Fixture managers provide a common context for fixtures
    across pytest and command-line environments.
    """

    plugin_name = 'nanaimo'

    type_factory_spec = pluggy.HookspecMarker(plugin_name)
    type_factory = pluggy.HookimplMarker(plugin_name)

    def __init__(self) -> None:
        self._pluginmanager = pluggy.PluginManager(self.plugin_name)
        self._fixture_cache = dict()  # type: typing.Dict[str, Fixture]

        class PluginNamespace:
            @self.type_factory_spec
            def get_fixture_type(self) -> typing.Type['Fixture']:
                raise NotImplementedError()

        self._pluginmanager.add_hookspecs(PluginNamespace)
        self._pluginmanager.load_setuptools_entrypoints(self.plugin_name)

    def fixture_types(self) -> typing.Generator:
        """
        Yields each fixture type registered with this object. The types may or may not
        have already been instantiated.
        """
        for fixture_type in self._pluginmanager.hook.get_fixture_type():
            yield fixture_type

    def get_fixture(self, fixture_name: str) -> Fixture:
        """
        Get a fixture instance creating it if it wasn't already.

        :param fixture_name: The canonical name of the fixture.
        :type fixture_name: str
        :param loop:
        :return: A manager-scoped fixture instance (i.e. One-and-only-one
            fixture instance with this name for this manager object).
        """
        try:
            return self._fixture_cache[fixture_name]
        except KeyError:
            for fixture_type in self._pluginmanager.hook.get_fixture_type():
                if fixture_type.get_canonical_name() == fixture_name:
                    fixture = typing.cast(Fixture, fixture_type(self))
                    self._fixture_cache[fixture_name] = fixture
                    return fixture
            raise



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
import functools
import typing

import pytest

import nanaimo
import nanaimo.pytest_plugin


class Fixture(nanaimo.Fixture):
    """
    A trivial plugin. Returns an artifact named "eat" that will log a yummy info message.
    """

    fixture_name = 'nanaimo_bar'

    @classmethod
    def on_visit_test_arguments(cls, arguments: nanaimo.Arguments) -> None:
        pass

    async def gather(self, args: nanaimo.Namespace) -> nanaimo.Artifacts:
        """
        Create a delicious function in the artifacts to eat.
        """
        artifacts = nanaimo.Artifacts()
        self.logger.info("don't forget to eat your dessert.")
        setattr(artifacts, 'eat', functools.partial(self.logger.info, 'Nanaimo bars are yummy.'))
        return artifacts


@nanaimo.FixtureManager.type_factory
def get_fixture_type() -> typing.Type['nanaimo.Fixture']:
    return Fixture


@pytest.fixture
def nanaimo_bar(request: typing.Any) -> nanaimo.Fixture:
    return nanaimo.pytest_plugin.create_pytest_fixture(request, Fixture)



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
"""
Built-in :class:`Fixture` objects for common scenarios. See individual fixture documentation for use.
"""
